!--------------------------------------------------------------------
subroutine md1dtest(uri,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!         model = 1; 1D Test for debugging

  use pram, only : imax, jmax, kmax, kmax, nv, gam, c0
  implicit none

  integer :: i, j, k, is1, ie1, js1, je1, ks1, ke1
  integer :: merr

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: de, pr
  real(8), allocatable :: util(:), bcon(:)
!
!-allocate variables -!
  allocate(util(1:3), bcon(1:3), stat=merr)  

!--------------------------------------------------------------------
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
  
!        gam=2.d0
          de=1.d0
          pr=0.01d0
          util(1)=0.d0
          util(2)=0.d0
          util(3)=0.d0
          bcon(1)=0.d0
          bcon(2)=0.d0 
          bcon(3)=0.0d0
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
          uri(6,i,j,k)=0.d0
          uri(7,i,j,k)=bcon(1)
          uri(8,i,j,k)=bcon(2)          
          uri(9,i,j,k)=bcon(3)
!
      enddo
    enddo
  enddo
!
  deallocate(util, bcon, stat=merr)
!  
  return
end subroutine md1dtest
!
!--------------------------------------------------------------------
subroutine mdshoctub(uri,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!         model = 2; 1D Shock Tube Test Problem

  use pram, only : imax, jmax, kmax, kmax, nv, gam, c0
  implicit none

  integer :: i, j, k, is1, ie1, js1, je1, ks1, ke1
  integer :: merr

    real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: de, pr
  real(8), allocatable :: util(:), bcon(:)
!
!-allocate variables -!
  allocate(util(1:3), bcon(1:3), stat=merr)   

!--------------------------------------------------------------------
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
  
!        gam=2.d0
        if(i .le. imax/2) then
          de=1.d0
          pr=1.d0
          util(1)=0.d0
          util(2)=0.d0
          util(3)=0.d0
          bcon(1)=0.5d0
          bcon(2)=1.0d0 
          bcon(3)=0.0d0
        else
          de=0.125d0
          pr=0.1d0
          util(1)=0.d0
          util(2)=0.d0
          util(3)=0.d0
          bcon(1)=0.5d0
          bcon(2)=-1.0d0
          bcon(3)=0.0d0
        endif
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
          uri(6,i,j,k)=0.d0
          uri(7,i,j,k)=bcon(1)
          uri(8,i,j,k)=bcon(2)          
          uri(9,i,j,k)=bcon(3)
!
      enddo
    enddo
  enddo
!
  deallocate(util, bcon, stat=merr)
!  
  return
end subroutine mdshoctub
!--------------------------------------------------------------------
subroutine md1dacc(uri,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!         model = 1; 1D GRHD Test for debugging (free-fall)

  use pram, only : imax, jmax, kmax, kmax, nv, gam, c0
  implicit none

  integer :: i, j, k, is1, ie1, js1, je1, ks1, ke1
  integer :: merr
!
  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: de, pr
  real(8), allocatable :: util(:), bcon(:)
!
!-allocate variables -!
  allocate(util(1:3), bcon(1:3), stat=merr)  
!
!--------------------------------------------------------------------
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
  
!        gam=2.d0
          de=1.d-1
          pr=1.d-3
          util(1)=0.d0
          util(2)=0.d0
          util(3)=0.d0
          bcon(1)=0.d0
          bcon(2)=0.d0 
          bcon(3)=0.0d0
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
          uri(6,i,j,k)=0.d0
          uri(7,i,j,k)=bcon(1)
          uri(8,i,j,k)=bcon(2)          
          uri(9,i,j,k)=bcon(3)
!
      enddo
    enddo
  enddo

  deallocate(util, bcon, stat=merr)  
  
  return
end subroutine md1dacc
!
!--------------------------------------------------------------------
subroutine mdeqcor1(uri,x1,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!         model = 1; 1D GRHD Test for debugging (free-fall)

  use pram, only : imax, jmax, kmax, kmax, nv, gam
  implicit none

  integer :: i, j, k, is1, ie1, js1, je1, ks1, ke1
  integer :: merr

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: de, pr
  real(8), allocatable :: util(:), bcon(:)
  real(8) :: x1(imax)
  real(8) :: rdcen1, ck1, ee0, dd0, cssq, rr
!
!-allocate variables -!
  allocate(util(1:3), bcon(1:3), stat=merr)  
!
!--------------------------------------------------------------------
! Parameter
!
  rdcen1=10.0d0
  ck1=0.1d0
  ee0=0.1d0
  dd0=1.d0

  cssq=gam*(gam-1.0)*ee0/dd0
!
!--------------------------------------------------------------------
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1  

        rr=x1(i)
        de=dd0*exp(-3.0*rr/rdcen1)

!        pr=ck1*de**gam
        pr=ck1*de**((gam+1.0)/gam)
!        pr=cssq/gam*de
!        pr=(gam-1.0)*e0*de

        util(1)=0.d0
        util(2)=0.d0
        util(3)=0.d0

        bcon(1)=0.d0
        bcon(2)=0.d0
        bcon(3)=0.d0
!
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!        
      enddo
    enddo
  enddo
!
  deallocate(util, bcon, stat=merr)
!
  return
end subroutine mdeqcor1
!
!--------------------------------------------------------------------
subroutine mdeqcor2(uri,x1,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!         model = 1; 1D GRHD Test for debugging (free-fall)

  use pram, only : imax, jmax, kmax, kmax, nv, gam
  implicit none

  integer :: i, j, k, is1, ie1, js1, je1, ks1, ke1
  integer :: merr
!  
  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: de, pr
  real(8), allocatable :: util(:), bcon(:)
  real(8) :: x1(imax)
!
  real(8) :: dd0, rr, ee0, pp0, rin
!
!-allocate variables -!
  allocate(util(1:3), bcon(1:3), stat=merr)
!  
!--------------------------------------------------------------------
! Parameter
!
  dd0=0.0001d0
  pp0=0.01d0*dd0
  rin=1.d0
!
!--------------------------------------------------------------------
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1  

        rr=x1(i)
        de=dd0*(rr/rin)**(-1.5d0)
!        ee0=0.1*dd0*rr**(-5./2.)
        ee0=pp0*(rr/rin)**(-2.5d0)
!        pr=(gam-1.0)*ee0
        pr=(gam-1.d0)*pp0*(rr/rin)**(-2.5d0)

!        util(1)=0.d0
        util(1)=-1.d0/(rr*rr)
        util(2)=0.d0
        util(3)=0.d0

        bcon(1)=0.d0
        bcon(2)=0.d0
        bcon(3)=0.d0
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!        
      enddo
    enddo
  enddo
!
  deallocate(util, bcon, stat=merr)
!    
  return
end subroutine mdeqcor2
!
!--------------------------------------------------------------------
subroutine mdffcor1(uri,gcov,gcon,x1,x2,x3,x1b,x2b,x3b, &
                    is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!   Bondi Accretion Model based on Hawley, Smarr, Wilson (1984)
!   + angular momentum (Proga et al. 2003)
!
  use pram, only : imax, jmax, kmax, nv, gam, c0, rbh, metric, akm, pi
  implicit none

  integer :: i, j, k, m, n, is1 ,ie1, js1, je1, ks1, ke1
  integer :: merr, iflg1, irs, ichk

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: gcov(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: gcon(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
    
  real(8) :: de, pr

  real(8) :: x1(imax), x2(jmax), x3(kmax) 
  real(8) :: x1b(imax), x2b(jmax), x3b(kmax)

  real(8), allocatable :: util(:), bcon(:)
  real(8), allocatable :: gcov1(:,:), gcon1(:,:)
  real(8), allocatable :: ucov(:), ucon(:), bbcov(:), bbcon(:)
  real(8), allocatable :: ffur(:), ffte(:)
  real(8), allocatable :: utiln(:)

  real(8) :: dd0, vp, rc, gam1, urcsq, urc, vvcsq, tec, cnst1, cnst2, &
             rr, qt, ur, te, al0, ffunc, uphi, b0, tmp1, tmp1a, tmp1b, &
              gfl, bbsq, beta, beta_act, bnorm
  real(8) :: x1aa, x2aa, x3aa, x3t   
!
!-allocate variables -!
  allocate(util(1:3), bcon(1:3), stat=merr)
  allocate(gcov1(0:3,0:3), gcon1(0:3,0:3), stat=merr)
  allocate(ucov(0:3), ucon(0:3), bbcov(0:3), bbcon(0:3), stat=merr)
  allocate(ffur(imax), ffte(imax), utiln(1:3), stat=merr)
!
! 
!--------------------------------------------------------------------
!  Parameter
!
  gam1=1.d0/(gam-1.d0)
!
  dd0=1.0d0 !- normalized density -!
  vp=1.d0 !- velocity correction -!
  rc=8.d0 !- sonic point radius -!
  al0=0.0d0 !- angular momentum paramter -!
!  b0=0.5084d0 !- magnetic field strength parameter -!
  b0=0.0d0
  
!
  urcsq=1./(2.*rc) !- (u^r_c)^2
  urc=-sqrt(urcsq) !
  vvcsq=(urcsq)/(1.-3.*urcsq)
!  tec=gam1*vvcsq/((1.+gam1)*(1.-gam1*vvcsq))
  tec=gam1*vvcsq/((1.+gam1)-gam1*(1.+gam1)*vvcsq)
! 
  cnst1=(tec**gam1)*urc*(rc**2)
!  cnst2=(1.-(2.*rbh/rc)+urcsq)*(1.+(1.+gam1)*tec)**2
  tmp1a=cnst1**2
  tmp1b=(rc**4)*(tec**(2.*gam1))
  tmp1=tmp1a/tmp1b
  cnst2=(1.-(2./rc)+tmp1)*((1.+(1.+gam1)*tec)**2)
!
  iflg1=0
  irs=1
!      
  do i=1,imax
    if(iflg1 .eq. 0) then
      if(x1(i) .ge. rc) then
        iflg1=1
        irs=i
      endif
    endif
  enddo
!
!======================================================================@
!     Calculation of alpha and ur by Newton-Raphson
!======================================================================@
!
  do i=1,imax
    ffur(i)=urc
    ffte(i)=tec
  enddo
!
! ==== rr > rcs ====  
!
  do i=irs,imax,1
  
    rr=x1(i)
    
    if(i .gt. irs) then
      ur=ffur(i-1)+0.01
      te=ffte(i-1)
    else
      ur=urc
      te=tec
    endif
       
    call calws1(rr,ur,te,cnst1,cnst2)
    
    ffur(i)=ur
    ffte(i)=te

    
  enddo
!
! ==== rr < rcs ====
!
  do i=irs-1,1,-1

    rr=x1(i)

    if(i .eq. irs-1) then
      ur=urc
      te=tec
    else
      ur=ffur(i+1)-0.01
      te=ffte(i+1)
    endif

    call calws1(rr,ur,te,cnst1,cnst2)

    ffur(i)=ur
    ffte(i)=te

  enddo
!
  ichk=0
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1 
 
        rr=x1(i)
        qt=x3(k)
!
        de=dd0*ffte(i)**gam1
        pr=de*ffte(i)
        util(1)=vp*ffur(i)
        util(2)=0.d0
        util(3)=0.d0
        bcon(1)=b0/rr**2
        bcon(2)=0.d0
        bcon(3)=0.d0          
!
!- Angular momentum -!
!
!        ffunc=1.- abs(cos(qt))
!        ffunc=1.- (cos(qt)**10.0)
!        if(qt .le. theta0 .or. qt .ge. pi-theta0)
!          ffunc=0.d0
!        else
!          ffunc=1.d0
!        endif
!
!        uphi=al0*ffunc/rr 
!        util(2)=uphi                
!
!- transform from BL coordinate to KS coordinate -!        
        if(metric .eq. 303 .or. metric .eq. 403) then
!
          x1aa=x1(i)
          x2aa=x2(j)
          x3aa=x3(k)   
          x3t=x3b(k)
!
          call transbl2ks(util,utiln,x1aa,x2aa,x3aa,x3t)
!
          util(1)=utiln(1)
          util(2)=utiln(2)
          util(3)=utiln(3)
!
        endif
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!       
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
            gcon1(m,n)=gcon(m,n,i,j,k)
          enddo
        enddo
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(rr .gt. rc .and. ichk .eq. 0 .and. j .eq. 4 .and. k .eq. 4) then
          beta_act=2.*pr/bbsq 
          write(*,*) 'beta, bb/ro=', beta_act, bbsq/de
          ichk=1 
         endif
!       
!        if(bbsq .gt. bsq_max) then
!          bsq_max=bbsq
!        endif
!        
      enddo
    enddo
  enddo
!
  deallocate(gcov1, gcon1, ucov, ucon, bbcov, bbcon, &
             ffur,ffte, util, bcon, utiln, stat=merr)
 
  return
end subroutine mdffcor1
!
!-----------------------------------------------------------------------
subroutine calws1(rr,ur,te,cnst1,cnst2)
!----------------------------------------------------------------------- 
  use pram, only : gam, c0, rbh, iter
  implicit none

  integer :: nnn
  real(8) :: rr, ur, te, cnst1, cnst2, gam1
  real(8) :: f1, g1, dfx, dfy, dgx, dgy, det, dx1, dy1, delta, deltaend
  real(8) :: ursq, rrsq, tmp3a, tmp3b, xx_old, yy_old, tmp1a, tmp1b 

!     Parameter for Bondi Accretion

  delta=1.d0
  deltaend=1.d-3
  gam1=1.d0/(gam-1.d0)

  do nnn=0,iter
    rrsq=rr*rr
    ursq=ur*ur

    f1=ur*(te**gam1)*rrsq-cnst1
    tmp3a=(1.+(1.+gam1)*te)**2
    tmp3b=1.-(2.*rbh/rr)+ursq
    tmp1a=cnst1**2
    tmp1b=(rr**4)*te**(2.*gam1)
!    tmp3b=1.-(2.*rbh/rr)+tmp1a/tmp1b
    g1=tmp3a*tmp3b-cnst2
      
    dfx=(te**gam1)*rrsq
    dfy=ur*gam1*(te**(gam1-1.))*rrsq
    dgx=2.*tmp3a*ur
!    dgx=0.d0
    dgy=2.*(1.+gam1)*(1.+(1.+gam1)*te)*tmp3b
!    dgy=2.*(1.+gam1)*(1.+(1.+gam1)*te)*tmp3b &
!       -(2.*gam1*(cnst1**2)/(rr**4))*tmp3a*te**(-2.*gam1-1.)

    det=dfx*dgy-dfy*dgx

!      write(*,*) 'f1, g1=', f1, g1
!      write(*,*) 'dfx,dfy,dgx,dgy=', dfx, dfy, dgx, dgy

    if(det .eq. 0.d0) then
      write(6,*) ' >> Jacobian =0 in grbondi'
      write(6,*) ' >> delta, iter =', delta, nnn,', at x1:',rr
 !       stop
    endif

    if(det .eq. 0.d0) then
      dx1=0.d0
      dy1=0.d0
      delta=0.d0
    else        
      dx1=(-dgy*f1+dfy*g1)/det
      dy1=(dgx*f1-dfx*g1)/det

      delta=abs((dx1+dy1)/2.0)
    endif

    if(delta .lt. deltaend) then
      goto 101
    endif

    xx_old=ur
    yy_old=te

    ur=ur+dx1
    te=te+dy1

  enddo

  if(delta .gt. deltaend) then 
    write(6,*) ' >> Not convergence in grbondi'
    write(6,*) ' >> delta, iter =', delta, nnn,', at x1:',rr
    stop
  endif
!
!----------------------------------------------------
!     Calculation of primitive variables
!----------------------------------------------------
!
  101 continue 

  return
end subroutine calws1
!
!--------------------------------------------------------------------
subroutine callfish(rmax,clm)
!--------------------------------------------------------------------
  use pram, only : akm
  implicit none

  real(8) :: rmax, clm
  real(8) :: tmp1a, tmp1b, tmp1c, tmp1d, tmp1e, tmp1f, &
             tmp2a, tmp2b, tmp3a, tmp3b
  tmp1a=((akm**2)-2.*akm*sqrt(rmax)+ (rmax**2))
  tmp1b=(-2.*akm*rmax)*tmp1a
  tmp1c=sqrt(2.*akm*sqrt(rmax)+(-3.+rmax)*rmax)
  tmp1d=(akm+(-2.+rmax)*sqrt(rmax))
  tmp1e=((rmax**3) + (akm**2)*(2.+rmax))
  tmp1f=sqrt(1.+(2.*akm/rmax**(1.5))-(3./rmax))

  tmp2a=(rmax**3)*sqrt(2.*akm*sqrt(rmax)+(-3.+rmax)*rmax)
  tmp2b=((akm**2) + (-2.+rmax)*rmax)

  tmp3a=tmp1a*((tmp1b/tmp1c)+tmp1d*tmp1e/tmp1f)
  tmp3b=tmp2a*tmp2b

  clm=tmp3a/tmp3b

  return
end subroutine callfish

!--------------------------------------------------------------------
subroutine mdtorus1(uri,gcov,gcon,detg, &
                    x1,x2,x3,x1b,x2b,x3b,dx1,dx2,dx3, &
                    myrank,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!
  use pram, only : imax, jmax, kmax, nv, gam, metric, akm, pi, kpol
  implicit none
  include 'mpif.h'

  integer :: i, j, k, m, n, is1 ,ie1, js1, je1, ks1, ke1
  integer :: merr, myrank, jh
   
  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: gcov(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: gcon(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
!
  real(8) :: detg(is1:ie1,js1:je1,ks1:ke1)
!
  real(8), allocatable :: gcov1(:,:), gcon1(:,:)
  real(8), allocatable :: ucov(:), ucon(:), bbcov(:), bbcon(:), beta1(:)
  real(8), allocatable :: util(:), utiln(:), bcon(:)
!
  real(8) :: x1(imax), x2(jmax), x3(kmax) 
  real(8) :: x1b(imax), x2b(jmax), x3b(kmax) 
  real(8) :: dx1(imax), dx2(jmax), dx3(kmax)
!
  real(8),allocatable :: alnh(:,:,:), pom(:,:,:)
!
  real(8) :: de, pr
  real(8) :: rin, rmax, kappa, beta, rhomin, prmin, clm, hm1, &
             expm2chi, uphi1, rhomax1, rhomax1a, &
             prmax1, prmax1a, prmax2, prmax2a, rin1,thedge, thedge1, &
             rhomin1, prmin1
  real(8) :: rr, th, sth, cth, del, aa, sig, rr1, th1, &
             thin, sthin, cthin, delin, aain, sigin
  real(8) :: x1aa, x2aa, x3aa, x3t
  real(8) :: tmp1a, tmp1b, tmp1c, tmp1d, tmp1e, tmp1f, &
             tmp2a, tmp2b, tmp2c, tmp2d, tmp2e, tmp2f, &
             tmp3a, tmp3b, tmp3c, tmp3d, tmp_rand
  real(8) :: bsq_max, bsq_max1, bsq_max2, bsq_max2a, beta_act, &
             bnorm, pom1, rhoav, gfl, bbsq
  real(8) :: tmp4a, tmp4b, tmp4c, discr, alpha1
! 
!- allocate variables -!
  allocate(gcov1(0:3,0:3), gcon1(0:3,0:3), stat=merr)
  allocate(util(1:3), utiln(1:3), bcon(1:3), ucov(0:3), ucon(0:3), &
           bbcov(0:3), bbcon(0:3), beta1(1:3), stat=merr)
  allocate(alnh(is1:ie1,js1:je1,ks1:ke1), pom(is1:ie1,js1:je1,ks1:ke1), &
           stat=merr)

!--------------------------------------------------------------------
!- Parameter -!

!  rin=6.d0
  rin=8.0d0
!  rmax=12.d0
!  rmax=13.d0
  rmax=15.0d0
  thedge=0.5d0
  thedge1=2.*pi-0.5d0

  kappa=1.0d-3
  beta=1.d2
!
  rin1=1.d0 
!  rhomin=5.d-4
  rhomin=5.d-5
  prmin=1.d-3*rhomin
!
  rhomax1=0.d0
  prmax1=0.d0 
  prmax2=0.d0
!- calculate Amgular momentum l -!
  call callfish(rmax,clm)
  if(myrank .eq. 0) then
    write(*,*) 'rmax, clm=', rmax, clm
  endif
!
!- initally set minimum plasma everywhere -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1

        rr=x1(i)
!
!        de=rhomin*((rr/rin1)**(-1.5d0))
!        de=rhomin*((rr/rin1))**(-1.0d0)
!        pr=(gam-1.d0)*prmin*((rr/rin1)**(-2.5d0))
!        pr=1.0d-3*de**gam
!        pr=1.0d-3*(gam-1.)*de
!
        de=rhomin
        pr=prmin
!
        util(1)=0.d0
!        util(1)=-1.d0/(rr*rr)
        util(2)=0.d0
        util(3)=0.d0
        bcon(1)=0.d0
        bcon(2)=0.d0 
        bcon(3)=0.d0
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!
        alnh(i,j,k)=0.d0
!        
      enddo
    enddo
  enddo
!
!- calculate ln h -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        rr=x1(i)
        th=x3(k)
        sth=sin(th)
        cth=cos(th) 
!
        del=(rr*rr)-(2.*rr)+(akm*akm)
        aa=(rr*rr+akm*akm)*(rr*rr+akm*akm)-del*akm*akm*sth*sth
        sig=rr*rr+akm*akm*cth*cth
!
        if(rr .ge. rin .and. th .ge. thedge .and. th .le. thedge1) then
!
          thin=pi/2.
          sthin=sin(thin)
          cthin=cos(thin)
          delin=rin*rin-2.*rin+akm*akm
          aain=(rin*rin+akm*akm)*(rin*rin+akm*akm)-delin*akm*akm*sthin*sthin
          sigin=rin*rin+akm*akm*cthin*cthin   
           
          tmp1a=4.*(clm*clm*sig*sig)*del/(aa*sth*aa*sth)
!          tmp1b=aa*sth*aa*sth
          tmp1c=sig*del/aa
          tmp1d=0.5*log((1.+sqrt(1.+tmp1a))/tmp1c)

          tmp1e=0.5*sqrt(1.+tmp1a)
          tmp1f=2.*akm*rr*clm/aa

          tmp2a=4.*(clm*clm*sigin*sigin)*delin/(aain*sthin*aain*sthin)
!          tmp2b=aain*sthin*aain*sthin
          tmp2c=sigin*delin/aain
          tmp2d=0.5*log((1.+sqrt(1.+tmp2a))/tmp2c)

          tmp2e=0.5*sqrt(1.+tmp2a)
          tmp2f=2.*akm*rin*clm/aain

          alnh(i,j,k)=tmp1d-tmp1e-tmp1f-(tmp2d-tmp2e-tmp2f)
          if(alnh(i,j,k) .le. 0.d0) then
            alnh(i,j,k)=0.d0 
          endif   
!
        else
          alnh(i,j,k)=0.d0
        endif
!
!- set torus structure -!
!
        hm1=exp(alnh(i,j,k)) -1.        
        if(hm1 .gt. 1.d-4 .and. rr .ge. rin &
           .and. th .ge. thedge .and. th .le. thedge1) then    
!    
!!          hm1=exp(alnh(i,j,k)) -1.
          de=(hm1*(gam-1.)/(kpol*gam))**(1./(gam-1.))
          pr=kpol*de**gam
!
!          de=rhomin*((rr/rin1)**(-1.5d0))
!          de=rhomin*((rr/rin1))**(-1.0d0)
!          pr=(gam-1.d0)*prmin*((rr/rin1)**(-2.5d0))

          util(1)=0.d0
          util(3)=0.d0
!
          expm2chi=sig*sig*del/(aa*aa*sth*sth)
          uphi1=sqrt((-1.+sqrt(1.+4.*clm*clm*expm2chi))/2.)
          tmp3a=2.*akm*rr*sqrt(1.+uphi1*uphi1)
          tmp3b=sqrt(aa*sig*del)
          tmp3c=sqrt(sig/aa)*uphi1/sth
          tmp3d=tmp3a/tmp3b+tmp3c
          util(2)=tmp3d
!          util(2)=1.03d0*tmp3d   
!          util(2)=0.d0
!
!- Put perturbation in pressure -!
!
          call random_number(tmp_rand)
!!          write(*,*) "tmp_rand=", tmp_rand-0.5
          pr=pr*(1.+0.03*(tmp_rand-0.5))
!
!- check density & pressure max -!
!
          if(de .gt. rhomax1) then
            rhomax1=de
          endif
          if(pr .gt. prmax1) then
            prmax1=pr
          endif
!
!- transform velocity from BL to KS coordinates -!
!
          if(metric .eq. 303 .or. metric .eq. 403) then
            x1aa=x1(i)
            x2aa=x2(j)
            x3aa=x3(k)   
            x3t=x3b(k)
            call transbl2ks(util,utiln,x1aa,x2aa,x3aa,x3t)
            util(1)=utiln(1)
            util(2)=utiln(2)
            util(3)=utiln(3)
!
          elseif(metric .eq. 203 .and. akm .ne. 0.d0) then

            do m=0,3
              do n=0,3
                gcov1(m,n)=gcov(m,n,i,j,k)
                gcon1(m,n)=gcon(m,n,i,j,k)
              enddo
            enddo

            alpha1=1./sqrt(-gcon1(0,0))
!
            beta1(1)=alpha1*alpha1*gcon1(0,1)
            beta1(2)=alpha1*alpha1*gcon1(0,2)
            beta1(3)=alpha1*alpha1*gcon1(0,3)
!
            ucon(1)=util(1)
            ucon(2)=util(2)
            ucon(3)=util(3)
!
            tmp4a=gcov1(0,0)
            tmp4b=2.*(gcov1(0,1)*ucon(1)+gcov1(0,2)*ucon(2)+gcov1(0,3)*ucon(3))
            tmp4c=1.+gcov1(1,1)*ucon(1)*ucon(1)+gcov1(2,2)*ucon(2)*ucon(2) &    
                  +gcov1(3,3)*ucon(3)*ucon(3) &
                  +2.*(gcov1(1,2)*ucon(1)*ucon(2)+gcov1(1,3)*ucon(1)*ucon(3) &  
                  +gcov1(2,3)*ucon(2)*ucon(3))   
            discr=tmp4b*tmp4b-4.*tmp4a*tmp4c
            ucon(0)=(-tmp4b-sqrt(discr))/(2.*tmp4a)
! 
            gfl=ucon(0)*alpha1
!
            util(1)=ucon(1)+gfl*beta1(1)/alpha1
            util(2)=ucon(2)+gfl*beta1(2)/alpha1
            util(3)=ucon(3)+gfl*beta1(3)/alpha1   
 
          endif
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
!          
        else
!          
!          de=rhomin*((rr/rin1)**(-1.5d0))
!          de=rhomin*((rr/rin1))**(-1.0d0)
!          pr=(gam-1.d0)*prmin*((rr/rin1)**(-2.5d0))
!          pr=1.0d-3*de(i,j,k)**gam
!
          de=rhomin
          pr=prmin
!
!          util(1)=0.d0
!          util(1)=-1.d0/(rr*rr)
!          util(2)=0.d0
!          util(3)=0.d0
!
          uri(1,i,j,k)=de
          uri(5,i,j,k)=pr
!
        endif
!        
      enddo
    enddo
  enddo
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(rhomax1,rhomax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  call mpi_allreduce(prmax1,prmax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
!
  if(myrank .eq. 0) then
    write(*,*) "rhomax1a, prmax1a=", rhomax1a, prmax1a
  endif
!
!
!- Normalize the densities and pressure -!
!
  rhomax1=0.d0
  prmax1=0.d0
  
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1 
        rr=x1(i)
        th=x3(k) 
!        rhomin1=rhomin*((rr/rin1)**(-1.5d0))
!        prmin1=(gam-1.d0)*prmin*((rr/rin1)**(-2.5d0))
        hm1=exp(alnh(i,j,k)) -1. 

        if(hm1 .gt. 1.d-4 .and. rr .ge. rin  &
             .and. th .ge. thedge .and. th .le. thedge1) then
!           
          de=uri(1,i,j,k)/rhomax1a
          pr=uri(5,i,j,k)/rhomax1a

!          de=uri(1,i,j,k)
!          pr=uri(5,i,j,k)
!!
!!          de=uri(1,i,j,k)/(1.d-1*rhomax1a)
!!          pr=uri(5,i,j,k)/(1.d-1*rhomax1a)
!!
          if(de .le. rhomin) then
            de=rhomin
          endif
          if(pr .le. prmin) then
            pr=prmin
          endif
!-check maximum density & pressure -!
          if(de .gt. rhomax1) then
            rhomax1=de
          endif
          if(pr .gt. prmax1) then
           prmax1=pr 
          endif

          uri(1,i,j,k)=de
          uri(5,i,j,k)=pr
!        else
!
!          de=rhomin1
!          de=rhomin*((rr/rin1))**(-1.0d0)
!          pr=prmin1
!          pr=1.0d-3*de(i,j,k)**gam 
         
        endif
!
      enddo
    enddo
  enddo
 !
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(rhomax1,rhomax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  call mpi_allreduce(prmax1,prmax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  
!   prmax2a=prmax1a/rhomax1a
!   rhomax1a=1.d0
!
  if(myrank .eq. 0) then
    write(*,*) 'rhomax1a, prmax1a=', rhomax1a, prmax1a
  endif
!
!
!- Set vector potential at cell-center (r-theta frame)-!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        pom(i,j,k)=0.d0
      enddo
    enddo
  enddo
!
  do k=ks1+1,ke1-1
    do j=js1,je1
      do i=is1+1,ie1-1
!
        rr1=x1(i)
        th1=x3(k)
!        rr1=(x1(i+1)+x1(i))/2.
!        th1=(x3(k+1)+x3(k))/2.
!        pom1=0.5*rr1*sin(th1)

!        rhoav=0.25*(uri(1,i-1,j,k)+uri(1,i,j,k-1) &
!                    +uri(1,i+1,j,k)+uri(1,i,j,k+1))
        rhoav=0.25*(uri(1,i,j,k)+uri(1,i-1,j,k) &
                    +uri(1,i,j,k-1)+uri(1,i-1,j,k-1))
!        rhoav=uri(1,i,j,k)
        pom1=rhoav/rhomax1a - 0.2
!       
        if(pom1 .gt. 0.d0) then
          pom(i,j,k)=pom1
        endif
!
      enddo
    enddo
  enddo
!
!- Set-up of Magnetic field -!
  do k=ks1+1,ke1-1
    do j=js1,je1
      do i=is1+1,ie1-1
        
!        bcon(1)=-(pom(i,j,k)-pom(i,j,k+1)+pom(i+1,j,k)-pom(i+1,j,k+1)) &
!                   /(2.*(x3(k+1)-x3(k))*detg(i,j,k))
!
!        bcon(3)=(pom(i,j,k)+pom(i,j,k+1)-pom(i+1,j,k)-pom(i+1,j,k+1)) &
!                   /(2.*(x1(i+1)-x1(i))*detg(i,j,k))
!
        if(pom(i,j,k) .gt. 0.d0) then         
          bcon(1)=-(pom(i,j,k)-pom(i,j,k+1)+pom(i+1,j,k)-pom(i+1,j,k+1)) &
                     /(2.*dx3(k)*detg(i,j,k))
          bcon(3)=(pom(i,j,k)+pom(i,j,k+1)-pom(i+1,j,k)-pom(i+1,j,k+1)) &
                    /(2.*dx1(i)*detg(i,j,k))
        endif 
!
!        if(k .eq. ks1) then
!          bcon(1)=(pom(i,j,k+1)-pom(i,j,k))/((x3(k+1)-x3(k))*detg(i,j,k))
!        elseif(k .eq. ke1) then
!          bcon(1)=(pom(i,j,k)-pom(i,j,k-1))/((x3(k)-x3(k-1))*detg(i,j,k))
!        else         
!          bcon(1)=(pom(i,j,k+1)-pom(i,j,k-1))/((x3(k+1)-x3(k-1))*detg(i,j,k))
!        endif
!
!        if(i .eq. is1) then
!          bcon(3)=-(pom(i+1,j,k)-pom(i,j,k))/((x1(i+1)-x1(i))*detg(i,j,k))
!        elseif(i .eq. ie1) then
!          bcon(3)=-(pom(i,j,k)-pom(i-1,j,k))/((x1(i)-x1(i-1))*detg(i,j,k))
!        else         
!          bcon(3)=-(pom(i+1,j,k)-pom(i-1,j,k))/((x1(i+1)-x1(i-1))*detg(i,j,k))
!        endif
!

!        if(j .eq. 4 .and. k .eq. 32) then
!           write(*,*) 'b1=', bcon1, i, j, k
!        endif
!
        uri(7,i,j,k)=bcon(1)
        uri(9,i,j,k)=bcon(3)
!        
      enddo
    enddo
  enddo
!
!- put boundary data for magnetic field -!
!
  do j=js1,je1
    do i=is1,ie1
      uri(7,i,j,ks1)=uri(7,i,j,ks1+1)
      uri(7,i,j,ke1)=uri(7,i,j,ke1-1)
!
      uri(9,i,j,ks1)=uri(9,i,j,ks1+1)
      uri(9,i,j,ke1)=uri(9,i,j,ke1-1)
    enddo
  enddo
!
  do k=ks1,ke1
    do j=js1,je1
      uri(7,is1,j,k)=uri(7,is1+1,j,k)
      uri(7,ie1,j,k)=uri(7,ie1-1,j,k)
!
      uri(9,is1,j,k)=uri(9,is1+1,j,k)
      uri(9,ie1,j,k)=uri(9,ie1-1,j,k)
    enddo
  enddo
!
!- calculation maximum magnetic field strength -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        util(1)=uri(2,i,j,k)
        util(2)=uri(3,i,j,k)
        util(3)=uri(4,i,j,k)
        bcon(1)=uri(7,i,j,k)
        bcon(2)=uri(8,i,j,k)
        bcon(3)=uri(9,i,j,k)
!       
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
            gcon1(m,n)=gcon(m,n,i,j,k)
          enddo
        enddo
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(bbsq .gt. bsq_max) then
          bsq_max=bbsq
        endif
!
      enddo
    enddo
  enddo    
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(bsq_max,bsq_max1,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
!
  beta_act=prmax1a/(0.5*bsq_max1)
!
  if(beta .eq. 0.d0) then
    bnorm=0.d0
  else
    bnorm=sqrt(beta_act/beta)
  endif
!
  if(myrank .eq. 0) then
    write(*,*) "beta_act, bnorm=", beta_act, bnorm
  endif
!
!- Normalize magnetic field strength -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        bcon(1)=uri(7,i,j,k)*bnorm            
        bcon(2)=uri(8,i,j,k)*bnorm  
        bcon(3)=uri(9,i,j,k)*bnorm  
!
        util(1)=uri(2,i,j,k)
        util(2)=uri(3,i,j,k)
        util(3)=uri(4,i,j,k)
!
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)
        uri(9,i,j,k)=bcon(3)
!       
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
            gcon1(m,n)=gcon(m,n,i,j,k)
          enddo
        enddo        
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(bbsq .gt. bsq_max2) then
          bsq_max2=bbsq
        endif
!
      enddo
    enddo
  enddo 
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(bsq_max2,bsq_max2a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  beta_act=prmax1a/(0.5*bsq_max2a)
!
  if(myrank .eq. 0) then
    write(*,*) "beta_act=", beta_act
  endif    
!
  deallocate(gcov1, gcon1, util, utiln, bcon, ucov, ucon, &
             bbcov, bbcon, beta1, alnh, pom, stat=merr)
!
  return
end subroutine mdtorus1
!
!--------------------------------------------------------------------
subroutine mdtorus2(uri,gcov,gcon,detg, &
                    x1,x2,x3,x1b,x2b,x3b,dx1,dx2,dx3, &
                    myrank,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!  power-law rotating torus model  (nearly Keplerian)
!  Devilliers et al. 2003, ApJ, 599, 1238
!
  use pram, only : imax, jmax, kmax, nv, gam, metric, akm, pi, iter, kpol
  implicit none
  include 'mpif.h'

  integer :: i, j, k, m, n, is1 ,ie1, js1, je1, ks1, ke1
  integer :: merr, myrank, jh, nnn

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  
  real(8) :: gcov(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: gcon(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
!
  real(8) :: detg(is1:ie1,js1:je1,ks1:ke1)
!
  real(8), allocatable :: gcov1(:,:), gcon1(:,:), gcon1in(:,:)
  real(8), allocatable :: ucov(:), ucon(:), bbcov(:), bbcon(:), beta1(:)
  real(8), allocatable :: util(:), utiln(:), bcon(:)
!
  real(8) :: x1(imax), x2(jmax), x3(kmax) 
  real(8) :: x1b(imax), x2b(jmax), x3b(kmax) 
  real(8) :: dx1(imax), dx2(jmax), dx3(kmax)
!
  real(8),allocatable :: ah(:,:,:), pom(:,:,:)
!
  real(8) :: de, pr  
  real(8) :: kappa, beta, clmin, clm, cqq, cqq1, eta, ckk, calp, calp2, &
             amdain, amda, utin, ut, uut, flin, fl, delta, deltaend, f, df, &
             rhomin, prmin, hm1, omg, thedge, thedge1, rcen, &
             rhomax1, rhomax1a, prmax1, prmax1a, prmax2, prmax2a 
  real(8) :: rr, th, sth, cth, del, aa, sig, rr1, th1, &
             rin, thin, sthin, cthin, delin, aain, sigin, rin1
  real(8) :: x1aa, x2aa, x3aa, x3t
  real(8) :: tmp_rand
  real(8) :: bsq_max, bsq_max1, bsq_max2, bsq_max2a, beta_act, &
             bnorm, pom1, rhoav, gfl, bbsq
  real(8) :: tmp4a, tmp4b, tmp4c, discr, alpha1
  real(8) :: tmp1, tmp5a1, tmp5a2, tmp5b1, tmp5b2, tmp5b3, tmp5b4 
! 
!- allocate variables -!
  allocate(gcov1(0:3,0:3), gcon1(0:3,0:3), gcon1in(0:3,0:3), stat=merr)
  allocate(util(1:3), utiln(1:3), bcon(1:3), ucov(0:3), ucon(0:3), &
           bbcov(0:3), bbcon(0:3), beta1(1:3), stat=merr)
  allocate(ah(is1:ie1,js1:je1,ks1:ke1), pom(is1:ie1,js1:je1,ks1:ke1), &
           stat=merr)

!--------------------------------------------------------------------
!- Parameter -!

  rin=15.d0
  thin=pi/2.d0
  kappa=1.0d-2
  beta=1.d2
!  beta=100.d0
!  beta=0.d0
!  clmin=4.57d0
  clmin=4.66d0
  cqq=1.68d0
  thedge=0.1d0
  thedge1=pi-0.1d0
  rcen=25.d0
!  
  rin1=1.d0
  rhomin=1.d-2 
  prmin=1.d-2*rhomin
!
  rhomax1=0.d0
  prmax1=0.d0 
  prmax2=0.d0
!
!- initally set minimum plasma everywhere -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1

        do m=0,3
          do n=0,3
            gcon1(m,n)=gcon(m,n,i,j,k)
          enddo
        enddo

        rr=x1(i)

        de=rhomin
!        de=rhomin*exp(-3.*rr/rcen)
!        de=rhomin*((rr/rin1)**(-1.5d0))

!        pr=(gam-1.d0)*prmin*((rr/rin1)**(-2.5d0))
!        pr=kpol*(de**gam)
!
        pr=prmin
        util(1)=0.d0
!        util(1)=(gcon1(0,1)/gcon1(0,0))*(1.-(1./rr)**4)
        util(2)=0.d0
        util(3)=0.d0
        bcon(1)=0.d0
        bcon(2)=0.d0 
        bcon(3)=0.d0
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!       
      enddo
    enddo
  enddo
!
!- calculation of constant paramter from lin  -!
!- calculation of BL Kerr metric at rin, thin -!
!  
  sthin=sin(thin)
  cthin=cos(thin)
  delin=rin*rin-2.*rin+akm*akm
  aain=(rin*rin+akm*akm)*(rin*rin+akm*akm)-delin*akm*akm*sthin*sthin
  sigin=rin*rin+akm*akm*cthin*cthin  
!
  do n=0,3
    do m=0,3
      gcon1in(m,n)=0.d0
    enddo
  enddo

  gcon1in(0,0)=-aain/(sigin*delin)               !- g^tt -!
  gcon1in(1,1)=delin/sigin                          !- g^rr -!
  gcon1in(2,2)=(sigin-2.*rin)/(delin*sigin*sthin*sthin) !- g^phi phi -!
  gcon1in(0,2)=-(2.*akm*rin)/(sigin*delin)           !- g^t phi -!
  gcon1in(2,0)=gcon1in(0,2)                  !- g^phi t -!
  gcon1in(3,3)=1./sigin                      !- g^th th -!
!
!- calculation of lambda at rin -!
  amdain=sqrt(clmin*(gcon1in(0,0)-clmin*gcon1in(0,2)) &
                 /(gcon1in(0,2)-clmin*gcon1in(2,2))) 
!- calculation of ut at rin -!
  tmp1=abs(gcon1in(0,0)-2.*clmin*gcon1in(0,2)+clmin*clmin*gcon1in(2,2))
  utin=-1./sqrt(tmp1) 
!- calculation of eta, k, alpha -!
  eta=clmin/(amdain**(2.-cqq))
  ckk=eta**(-2./(cqq-2.))
  calp=cqq/(cqq-2.)
  calp2=calp+1.
!- calculation of f(lin) -!
  flin=abs(1.-ckk*clmin**calp2)**(1./calp2)
!
  if(myrank .eq. 0) then
    write(*,*) 'lin, utin, flin, eta, ckk=', clmin, utin, flin, eta, ckk
  endif
!
!- calculate h -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        rr=x1(i)
        th=x3(k)
        sth=sin(th)
        cth=cos(th) 
!
        if(rr .ge. rin) then         
!        if(rr .ge. rin .and. th .ge. thedge .and. th .le. thedge1) then
!- calculation of BL Kerr metric -! 
          del=(rr*rr)-(2.*rr)+(akm*akm)
          aa=(rr*rr+akm*akm)*(rr*rr+akm*akm)-del*akm*akm*sth*sth
          sig=rr*rr+akm*akm*cth*cth
!
          do n=0,3
            do m=0,3
              gcon1(m,n)=0.d0
            enddo
          enddo
!
          gcon1(0,0)=-aa/(sig*del)               !- g^tt -!
          gcon1(1,1)=del/sig                          !- g^rr -!
          gcon1(2,2)=(sig-2.*rr)/(del*sig*sth*sth) !- g^phi phi -!
          gcon1(0,2)=-(2.*akm*rr)/(sig*del)           !- g^t phi -!
          gcon1(2,0)=gcon1(0,2)                  !- g^phi t -!
          gcon1(3,3)=1./sig                      !- g^th th -!        
!
!- calculation of l -!
          if(akm .eq. 0.d0) then
            amda=sqrt(-gcon1(0,0)/gcon1(2,2))
            clm=eta*amda**(2.-cqq)
            fl=abs(1.-ckk*clm**calp2)**(1./calp2)
          else
            delta=1.d0
            deltaend=0.0001d0
            f=1.d0
            df=1.d0
            nnn=0
!- initial value of l for Newton-Raphson method -!
            amda=sqrt(clmin*(gcon1(0,0)-clmin*gcon1(0,2)) &
                     /(gcon1(0,2)-clmin*gcon1(2,2)))
            clm=eta*amda**(2.-cqq)   
            cqq1=0.5*(2.-cqq)          
!
!- Newton Raphson iteration method -! 
  95        continue
!
            if(abs(delta) .gt. deltaend) then
!
!              tmp5a=clm*((gcon1(0,2)-clm*gcon1(2,2))**cqq1)
            
              tmp5a1=sqrt((clm*gcon1(0,0)-gcon1(0,2)*clm*clm) &
                     /(gcon1(0,2)-clm*gcon1(2,2)))
              tmp5a2=eta*tmp5a1**(2.-cqq)
              f=tmp5a2-clm
!
              tmp5b1=(gcon1(0,0)-2.*clm*gcon1(0,2)) &
                    *(gcon1(0,2)-clm*gcon1(2,2))
              tmp5b2=(clm*gcon1(0,0)-clm*clm*gcon1(0,2)) &
                    *(-gcon1(2,2))
              tmp5b3=(tmp5b1-tmp5b2)/(gcon1(0,2)-clm*gcon1(2,2))**2
              tmp5b4=tmp5a1**(-cqq)
              df=0.5*(2.-cqq)*eta*tmp5b3*tmp5b4-1.d0
!
              if(f .eq. 0.d0 .or. df .eq. 0.d0) then
                write(6,*) 'Bad numerical divergence in thick torus' 
                write(6,*) 'f, df =', f, df,'at', i, j, k
              else
                delta=f/df
              endif 
! 
              clm=clm-delta
              nnn=nnn+1
              if(nnn .gt. 1000) then
                write(6,*) 'iteration is over 1000 in thick torus'    
              endif

              goto 95
            endif
 
            fl=abs(1.-ckk*clm**calp2)**(1./calp2)          
          endif
!         
          tmp1=abs(gcon1(0,0)-2.*clm*gcon1(0,2)+clm*clm*gcon1(2,2))
          ut=-1./sqrt(tmp1) 

          ah(i,j,k)=(utin*flin)/(ut*fl)
!
          if(j .eq. 4 .and. i .eq. 60) then
             write(*,*) 'ah=', ah(i,j,k), 180.*x3(k)/pi
          endif

        else
          ah(i,j,k)=1.d0
        endif
!
!- set torus structure -!
!
        if(ah(i,j,k) .gt. 1.d0 .and. rr .ge. rin  &
           .and. th .ge. thedge .and. th .le. thedge1) then    
!    
          hm1=ah(i,j,k) -1.
          de=(hm1*(gam-1.)/(kpol*gam))**(1./(gam-1.))
          pr=kpol*de**gam
!
!         if(j .eq. 4 .and. k .eq. 32) then
!           write(*,*) 'ah, de=', ah(i,j,k), de(i,j,k), i
!         endif
!
          util(1)=0.d0
          util(3)=0.d0
!
          omg=(gcon1(0,2)-clm*gcon1(2,2))/(gcon1(0,0)-clm*gcon1(2,2))
          uut=-1.*ut/(1.-omg*clm)
!
          util(2)=(-omg*uut)*1.04d0   
!          util(2)=0.d0
!
!- Put perturbation in pressure -!
!
         call random_number(tmp_rand)
!!          write(*,*) "tmp_rand=", tmp_rand-0.5
          pr=pr*(1.+0.05*(tmp_rand-0.5))
!
!- check density & pressure max -!
!
          if(de .gt. rhomax1) then
            rhomax1=de
!            write(*,*) 'de,i,k,rr,th=', de,i,k,x1(i),x3(k)
          endif
          if(pr .gt. prmax1) then
            prmax1=pr
          endif
!
!- transform velocity from BL to KS coordinates -!
!
          if(metric .eq. 303 .or. metric .eq. 403) then
            x1aa=x1(i)
            x2aa=x2(j)
            x3aa=x3(k)   
            x3t=x3b(k)
            call transbl2ks(util,utiln,x1aa,x2aa,x3aa,x3t)
            util(1)=utiln(1)
            util(2)=utiln(2)
            util(3)=utiln(3)
!
          elseif(metric .eq. 203 .and. akm .ne. 0.d0) then
!
            do m=0,3
              do n=0,3
                gcov1(m,n)=gcov(m,n,i,j,k)
                gcon1(m,n)=gcon(m,n,i,j,k)
              enddo
            enddo
!
            alpha1=1./sqrt(-gcon1(0,0))
!
            beta1(1)=alpha1*alpha1*gcon1(0,1)
            beta1(2)=alpha1*alpha1*gcon1(0,2)
            beta1(3)=alpha1*alpha1*gcon1(0,3)
!
            ucon(1)=util(1)
            ucon(2)=util(2)
            ucon(3)=util(3)
!
            tmp4a=gcov1(0,0)
            tmp4b=2.*(gcov1(0,1)*ucon(1)+gcov1(0,2)*ucon(2)+gcov1(0,3)*ucon(3))
            tmp4c=1.+gcov1(1,1)*ucon(1)*ucon(1)+gcov1(2,2)*ucon(2)*ucon(2) &    
                  +gcov1(3,3)*ucon(3)*ucon(3) &
                  +2.*(gcov1(1,2)*ucon(1)*ucon(2)+gcov1(1,3)*ucon(1)*ucon(3) &  
                  +gcov1(2,3)*ucon(2)*ucon(3))   
            discr=tmp4b*tmp4b-4.*tmp4a*tmp4c
            ucon(0)=(-tmp4b-sqrt(discr))/(2.*tmp4a)
! 
            gfl=ucon(0)*alpha1
!
            util(1)=ucon(1)+gfl*beta1(1)/alpha1
            util(2)=ucon(2)+gfl*beta1(2)/alpha1
            util(3)=ucon(3)+gfl*beta1(3)/alpha1  
!
          endif
!
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
!          
        endif
!
      enddo
    enddo
  enddo
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(rhomax1,rhomax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  call mpi_allreduce(prmax1,prmax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)

  if(myrank .eq. 0) then
    write(*,*) "rhomax1a, prmax1a=", rhomax1a, prmax1a
  endif
!
!
!- Normalize the densities and pressure -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1 
        rr=x1(i)
        th=x3(k)
        if(ah(i,j,k) .gt. 1.d0 .and. rr .ge. rin &
           .and. th .ge. thedge .and. th .le. thedge1) then
          de=uri(1,i,j,k)/rhomax1a
          pr=uri(5,i,j,k)/rhomax1a
!
          if(de .le. rhomin) then
            de=rhomin
          endif
          if(pr .le. prmin) then
            pr=prmin
          endif
!
          uri(1,i,j,k)=de
          uri(5,i,j,k)=pr
        endif
!
      enddo
    enddo
  enddo
!
   prmax2a=prmax1a/rhomax1a
   rhomax1a=1.d0
!
  if(myrank .eq. 0) then
    write(*,*) 'rhomax1a, prmax2a=', rhomax1a, prmax2a
  endif
!
!
!- Set vector potential at cell-center (r-theta frame)-!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        pom(i,j,k)=0.d0
      enddo
    enddo
  enddo
!
  do k=ks1+1,ke1-1
    do j=js1,je1
      do i=is1+1,ie1-1
!
!        rr1=x1(i)
!        th1=x3(k)
!        rr1=(x1(i+1)+x1(i))/2.
!        th1=(x3(k+1)+x3(k))/2.
!        pom1=0.5*rr1*sin(th1)

!!        rhoav=0.25*(uri(1,i-1,j,k)+uri(1,i,j,k-1) &
!!                     +uri(1,i+1,j,k)+uri(1,i,j,k+1))
         rhoav=0.25*(uri(1,i,j,k)+uri(1,i-1,j,k) &
                     +uri(1,i,j,k-1)+uri(1,i-1,j,k-1))
!        rhoav=uri(1,i,j,k)
        pom1=rhoav/rhomax1a - 0.3
!
!         if(j .eq. 4 .and. k .eq. 64) then
!           write(*,*) 'pom1=', pom1, i, j, k
!         endif

        if(pom1 .gt. 0.d0) then
          pom(i,j,k)=pom1
        endif
!
      enddo
    enddo
  enddo
!
!- Set-up of Magnetic field -!
  do k=ks1+1,ke1-1
    do j=js1,je1
      do i=is1+1,ie1-1
        
!        bcon(1)=-(pom(i,j,k)-pom(i,j,k+1)+pom(i+1,j,k)-pom(i+1,j,k+1)) &
!                   /(2.*(x3(k+1)-x3(k))*detg(i,j,k))
!
!        bcon(3)=(pom(i,j,k)+pom(i,j,k+1)-pom(i+1,j,k)-pom(i+1,j,k+1)) &
!                   /(2.*(x1(i+1)-x1(i))*detg(i,j,k))

        bcon(1)=-(pom(i,j,k)-pom(i,j,k+1)+pom(i+1,j,k)-pom(i+1,j,k+1)) &
                   /(2.*dx3(k)*detg(i,j,k))
!
        bcon(3)=(pom(i,j,k)+pom(i,j,k+1)-pom(i+1,j,k)-pom(i+1,j,k+1)) &
                   /(2.*dx1(i)*detg(i,j,k))

!        if(k .eq. ks1) then
!          bcon(1)=(pom(i,j,k+1)-pom(i,j,k))/((x3(k+1)-x3(k))*detg(i,j,k))
!        elseif(k .eq. ke1) then
!          bcon(1)=(pom(i,j,k)-pom(i,j,k-1))/((x3(k)-x3(k-1))*detg(i,j,k))
!        else         
!          bcon(1)=(pom(i,j,k+1)-pom(i,j,k-1))/((x3(k+1)-x3(k-1))*detg(i,j,k))
!        endif
!
!        if(i .eq. is1) then
!          bcon(3)=-(pom(i+1,j,k)-pom(i,j,k))/((x1(i+1)-x1(i))*detg(i,j,k))
!        elseif(i .eq. ie1) then
!          bcon(3)=-(pom(i,j,k)-pom(i-1,j,k))/((x1(i)-x1(i-1))*detg(i,j,k))
!        else         
!          bcon(3)=-(pom(i+1,j,k)-pom(i-1,j,k))/((x1(i+1)-x1(i-1))*detg(i,j,k))
!        endif
!
        uri(7,i,j,k)=bcon(1)
        uri(9,i,j,k)=bcon(3)
!
      enddo
    enddo
  enddo
!
!- put boundary data for magnetic field -!
!
  do j=js1,je1
    do i=is1,ie1
      uri(7,i,j,ks1)=uri(7,i,j,ks1+1)
      uri(7,i,j,ke1)=uri(7,i,j,ke1-1)
!
      uri(9,i,j,ks1)=uri(9,i,j,ks1+1)
      uri(9,i,j,ke1)=uri(9,i,j,ke1-1)
    enddo
  enddo
!
  do k=ks1,ke1
    do j=js1,je1
      uri(7,is1,j,k)=uri(7,is1+1,j,k)
      uri(7,ie1,j,k)=uri(7,ie1-1,j,k)
!
      uri(9,is1,j,k)=uri(9,is1+1,j,k)
      uri(9,ie1,j,k)=uri(9,ie1-1,j,k)
    enddo
  enddo
!
!- calculation maximum magnetic field strength -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        util(1)=uri(2,i,j,k)
        util(2)=uri(3,i,j,k)
        util(3)=uri(4,i,j,k)
        bcon(1)=uri(7,i,j,k)
        bcon(2)=uri(8,i,j,k)
        bcon(3)=uri(9,i,j,k)
!       
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
          enddo
        enddo
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(bbsq .gt. bsq_max) then
          bsq_max=bbsq
        endif
!
      enddo
    enddo
  enddo    
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(bsq_max,bsq_max1,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)

  beta_act=prmax2a/(0.5*bsq_max1)
!
  if(beta .eq. 0.d0) then
    bnorm=0.d0
  else
    bnorm=sqrt(beta_act/beta)
  endif

  if(myrank .eq. 0) then
    write(*,*) "beta_act, bnorm=", beta_act, bnorm
  endif
!
!- Normalize magnetic field strength -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        bcon(1)=uri(7,i,j,k)*bnorm            
        bcon(2)=uri(8,i,j,k)*bnorm  
        bcon(3)=uri(9,i,j,k)*bnorm  
!
        util(1)=uri(2,i,j,k)
        util(2)=uri(3,i,j,k)
        util(3)=uri(4,i,j,k)
!
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)
        uri(9,i,j,k)=bcon(3)
!
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
          enddo
        enddo        
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(bbsq .gt. bsq_max2) then
          bsq_max2=bbsq
        endif
!
      enddo
    enddo
  enddo 
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(bsq_max2,bsq_max2a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  beta_act=prmax2a/(0.5*bsq_max2a)
!
  if(myrank .eq. 0) then
    write(*,*) "beta_act=", beta_act
  endif    
!
  deallocate(gcov1, gcon1, gcon1in, util, utiln, bcon, ucov, ucon, &
             bbcov, bbcon, beta1, ah, pom, stat=merr)
!
  return
end subroutine mdtorus2
!
!--------------------------------------------------------------------
subroutine mdtorus3(uri,gcov,gcon,detg, &
                    x1,x2,x3,x1b,x2b,x3b,dx1,dx2,dx3, &
                    myrank,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!  constant-l rotating torus model
!  Font & Daigne 2002, MNRAS, 334, 383
!
  use pram, only : imax, jmax, kmax, nv, gam, metric, akm, pi, iter, kpol
  implicit none
  include 'mpif.h'

  integer :: i, j, k, m, n, is1 ,ie1, js1, je1, ks1, ke1
  integer :: merr, myrank, jh, nnn

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  
  real(8) :: gcov(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: gcon(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
!
  real(8) :: detg(is1:ie1,js1:je1,ks1:ke1)
!
  real(8), allocatable :: gcov1(:,:), gcon1(:,:), gcov1in(:,:)
  real(8), allocatable :: ucov(:), ucon(:), bbcov(:), bbcon(:), beta1(:)
  real(8), allocatable :: util(:), utiln(:), bcon(:)
!
  real(8) :: x1(imax), x2(jmax), x3(kmax) 
  real(8) :: x1b(imax), x2b(jmax), x3b(kmax) 
  real(8) :: dx1(imax), dx2(jmax), dx3(kmax)
!
  real(8),allocatable :: ah(:,:,:), pom(:,:,:)
!
  real(8) :: de, pr  
  real(8) :: kappa, beta, clm, utin, ut, uut, &
             rhomin, prmin, hm1, omg, thedge, thedge1, rcen, &
             rhomax1, rhomax1a, prmax1, prmax1a, prmax2, prmax2a 
  real(8) :: rr, th, sth, cth, del, aa, sig, rr1, th1, &
             rin, thin, sthin, cthin, delin, aain, sigin, rin1
  real(8) :: x1aa, x2aa, x3aa, x3t
  real(8) :: tmp_rand
  real(8) :: bsq_max, bsq_max1, bsq_max2, bsq_max2a, beta_act, &
             bnorm, pom1, rhoav, gfl, bbsq
  real(8) :: tmp4a, tmp4b, tmp4c, discr, alpha1
  real(8) :: tmp1a, tmp1b, tmp2a, tmp2b
  real(8) :: ucovph
! 
!- allocate variables -!
  allocate(gcov1(0:3,0:3), gcon1(0:3,0:3), gcov1in(0:3,0:3), stat=merr)
  allocate(util(1:3), utiln(1:3), bcon(1:3), ucov(0:3), ucon(0:3), &
           bbcov(0:3), bbcon(0:3), beta1(1:3), stat=merr)
  allocate(ah(is1:ie1,js1:je1,ks1:ke1), pom(is1:ie1,js1:je1,ks1:ke1), &
           stat=merr)

!--------------------------------------------------------------------
!- Parameter -!

  rin=9.34d0
!  rin=9.0d0
  thin=pi/2.d0
  kappa=1.0d-3
!  beta=0.d0
  beta=100.d0
!  beta=0.d0
!  clm=4.5d0 ! a=0.0
!  clm=4.35d0 ! a=0.5
  clm=4.25d0 ! a=0.9
!  
  thedge=0.1d0
  thedge1=pi-0.1d0
  rcen=25.d0
!  
  rin1=1.d0
  rhomin=1.d-7 
  prmin=1.d-3*rhomin
!
  rhomax1=0.d0
  rhomax1a=0.d0
  prmax1=0.d0 
  prmax1a=0.d0
  prmax2=0.d0
!  
  bsq_max=0.d0
  bsq_max1=0.d0
  bsq_max2=0.d0
  bsq_max2a=0.d0
!
!- initally set minimum plasma everywhere -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1

        do m=0,3
          do n=0,3
            gcon1(m,n)=gcon(m,n,i,j,k)
          enddo
        enddo

        rr=x1(i)

        de=rhomin
!        de=rhomin*exp(-3.*rr/rcen)
!        de=rhomin*((rr/rin1)**(-1.5d0))

!        pr=(gam-1.d0)*prmin*((rr/rin1)**(-2.5d0))
!        pr=kpol*(de**gam)
!
        pr=prmin
        util(1)=0.d0
!        util(1)=(gcon1(0,1)/gcon1(0,0))*(1.-(1./rr)**4)
        util(2)=0.d0
        util(3)=0.d0
        bcon(1)=0.d0
        bcon(2)=0.d0 
        bcon(3)=0.d0
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!       
      enddo
    enddo
  enddo
!  
!- calculation of BL Kerr metric at rin, thin -!
!  
  sthin=sin(thin)
  cthin=cos(thin)
  delin=rin*rin-2.*rin+akm*akm
  aain=(rin*rin+akm*akm)*(rin*rin+akm*akm)-delin*akm*akm*sthin*sthin
  sigin=rin*rin+akm*akm*cthin*cthin  
!
  do n=0,3
    do m=0,3
      gcov1in(m,n)=0.d0
    enddo
  enddo

  gcov1in(0,0)=-(sigin-2.*rin)/sigin               !- g_tt -!
  gcov1in(1,1)=sigin/delin                          !- g_rr -!
  gcov1in(2,2)=(aain/sigin)*sthin*sthin !- g_phi phi -!
  gcov1in(0,2)=-(2.*akm*rin*sthin*sthin)/sigin  !- g_t phi -!
  gcov1in(2,0)=gcov1in(0,2)                  !- g_phi t -!
  gcov1in(3,3)=sigin                      !- g_th th -!
! 
!- calculation of ut at rin -!
  tmp1a=gcov1in(0,2)*gcov1in(0,2)-gcov1in(0,0)*gcov1in(2,2)
  tmp1b=gcov1in(0,0)*clm*clm+2.*clm*gcov1in(0,2)+gcov1in(2,2)
  if(tmp1b .gt. 0.d0) then
    utin=-sqrt(tmp1a/tmp1b)
  else
    utin=-1.d0  
  endif
!
  if(myrank .eq. 0) then
    write(*,*) 'lin, utin=', clm, utin
  endif
!
!- calculate h -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        rr=x1(i)
        th=x3(k)
        sth=sin(th)
        cth=cos(th) 
!
        if(rr .ge. rin) then         
!        if(rr .ge. rin .and. th .ge. thedge .and. th .le. thedge1) then
!- calculation of BL Kerr metric -! 
          del=(rr*rr)-(2.*rr)+(akm*akm)
          aa=(rr*rr+akm*akm)*(rr*rr+akm*akm)-del*akm*akm*sth*sth
          sig=rr*rr+akm*akm*cth*cth
!
          do n=0,3
            do m=0,3
              gcov1(m,n)=0.d0
              gcon1(m,n)=0.d0 
            enddo
          enddo
!
          gcov1(0,0)=-(sig-2.*rr)/sig               !- g^tt -!
          gcov1(1,1)=sig/del                          !- g^rr -!
          gcov1(2,2)=(aa/sig)*sth*sth             !- g^phi phi -!
          gcov1(0,2)=-(2.*akm*rr*sth*sth)/sig           !- g^t phi -!
          gcov1(2,0)=gcov1(0,2)                  !- g^phi t -!
          gcov1(3,3)=sig                      !- g^th th -!        
!
          gcon1(0,0)=-aa/(sig*del)               !- g^tt -!
          gcon1(1,1)=del/sig                          !- g^rr -!
          gcon1(2,2)=(sig-2.*rr)/(del*sig*sth*sth) !- g^phi phi -!
          gcon1(0,2)=-(2.*akm*rr)/(sig*del)           !- g^t phi -!
          gcon1(2,0)=gcon1(0,2)                  !- g^phi t -!
          gcon1(3,3)=1./sig 
!          
!- calculation of u_t at r,theta -!
!          
          tmp2a=gcov1(0,2)*gcov1(0,2)-gcov1(0,0)*gcov1(2,2)
          tmp2b=gcov1(0,0)*clm*clm+2.*clm*gcov1(0,2)+gcov1(2,2)
          if(tmp2b .gt. 0.d0) then
            ut=-sqrt(tmp2a/tmp2b)
          else
            ut=-1.d0  
          endif        

!
!- calculation of h = u_t,in/u_t
!          
          ah(i,j,k)=utin/ut
!
        else
          ah(i,j,k)=-1.d0
        endif
!
!- set torus structure -!
!
        if(ah(i,j,k) .gt. 1.d0 .and. rr .ge. rin  &
           .and. th .ge. thedge .and. th .le. thedge1) then    
!    
          hm1=max((ah(i,j,k) -1.)*(gam-1.)/gam, 0.d0)
          de=(hm1/kpol)**(1./(gam-1.))
          pr=kpol*(hm1/kpol)**(gam/(gam-1.))
!
!         if(j .eq. 4 .and. k .eq. 32) then
!           write(*,*) 'ah, de=', ah(i,j,k), de(i,j,k), i
!         endif
!
          util(1)=0.d0
          util(3)=0.d0
!
!- calculation of omega & contravariant u^t
!          
          omg=-(gcov1(0,2)+gcov1(0,0)*clm)/(gcov1(2,2)+gcov1(0,2)*clm)          
          uut=sqrt(-1./(gcov1(0,0)+2.*omg*gcov1(0,2)+omg*omg*gcov1(2,2)))
!          uut=-1./(ut*(1.-omg*clm))
!
          ucovph=-omg*ut

          util(2)=(omg*uut)
!          util(2)=gcon1(0,2)*ut+gcon1(2,2)*ucovph
!          util(2)=0.d0
!
!- Put perturbation in pressure -!
!
         call random_number(tmp_rand)
!!          write(*,*) "tmp_rand=", tmp_rand-0.5
           pr=pr*(1.+0.05*(tmp_rand-0.5))
           util(1)=util(1)+0.01*(tmp_rand-0.5)             
!
!- check density & pressure max -!
!
          if(de .gt. rhomax1) then
            rhomax1=de
!            write(*,*) 'de,i,k,rr,th=', de,i,k,x1(i),x3(k)
          endif
          if(pr .gt. prmax1) then
            prmax1=pr
          endif
!
!- transform velocity from BL to KS coordinates -!
!
          if(metric .eq. 303 .or. metric .eq. 403) then
            x1aa=x1(i)
            x2aa=x2(j)
            x3aa=x3(k)   
            x3t=x3b(k)
            call transbl2ks(util,utiln,x1aa,x2aa,x3aa,x3t)
            util(1)=utiln(1)
            util(2)=utiln(2)
            util(3)=utiln(3)
!
          elseif(metric .eq. 203 .and. akm .ne. 0.d0) then
!
            do m=0,3
              do n=0,3
                gcov1(m,n)=gcov(m,n,i,j,k)
                gcon1(m,n)=gcon(m,n,i,j,k)
              enddo
            enddo
!
            alpha1=1./sqrt(-gcon1(0,0))
!
            beta1(1)=alpha1*alpha1*gcon1(0,1)
            beta1(2)=alpha1*alpha1*gcon1(0,2)
            beta1(3)=alpha1*alpha1*gcon1(0,3)
!
            ucon(1)=util(1)
            ucon(2)=util(2)
            ucon(3)=util(3)
!
            tmp4a=gcov1(0,0)
            tmp4b=2.*(gcov1(0,1)*ucon(1)+gcov1(0,2)*ucon(2)+gcov1(0,3)*ucon(3))
            tmp4c=1.+gcov1(1,1)*ucon(1)*ucon(1)+gcov1(2,2)*ucon(2)*ucon(2) &    
                  +gcov1(3,3)*ucon(3)*ucon(3) &
                  +2.*(gcov1(1,2)*ucon(1)*ucon(2)+gcov1(1,3)*ucon(1)*ucon(3) &  
                  +gcov1(2,3)*ucon(2)*ucon(3))   
            discr=tmp4b*tmp4b-4.*tmp4a*tmp4c
            ucon(0)=(-tmp4b-sqrt(discr))/(2.*tmp4a)
! 
            gfl=ucon(0)*alpha1
!
            util(1)=ucon(1)+gfl*beta1(1)/alpha1
            util(2)=ucon(2)+gfl*beta1(2)/alpha1
            util(3)=ucon(3)+gfl*beta1(3)/alpha1  
!
          endif
!
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
!          
        endif
!
      enddo
    enddo
  enddo
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(rhomax1,rhomax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  call mpi_allreduce(prmax1,prmax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
!
  if(myrank .eq. 0) then
    write(*,*) "rhomax1a, prmax1a=", rhomax1a, prmax1a
  endif
!
!
!- Normalize the densities and pressure -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1 
        rr=x1(i)
        th=x3(k)
        if(ah(i,j,k) .gt. 1.d0 .and. rr .ge. rin &
           .and. th .ge. thedge .and. th .le. thedge1) then
          de=uri(1,i,j,k)/rhomax1a
          pr=uri(5,i,j,k)/rhomax1a
!
          if(de .le. rhomin) then
            de=rhomin
          endif
          if(pr .le. prmin) then
            pr=prmin
          endif
!
          uri(1,i,j,k)=de
          uri(5,i,j,k)=pr
        endif
!
      enddo
    enddo
  enddo
!
   prmax2a=prmax1a/rhomax1a
   rhomax1a=1.d0
!
  if(myrank .eq. 0) then
    write(*,*) 'rhomax1a, prmax2a=', rhomax1a, prmax2a
  endif
!
!
!- Set vector potential at cell-center (r-theta frame)-!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
        pom(i,j,k)=0.d0
      enddo
    enddo
  enddo
!
  do k=ks1+1,ke1-1
    do j=js1,je1
      do i=is1+1,ie1-1
!
!        rr1=x1(i)
!        th1=x3(k)
!        rr1=(x1(i+1)+x1(i))/2.
!        th1=(x3(k+1)+x3(k))/2.
!        pom1=0.5*rr1*sin(th1)

!!        rhoav=0.25*(uri(1,i-1,j,k)+uri(1,i,j,k-1) &
!!                     +uri(1,i+1,j,k)+uri(1,i,j,k+1))
         rhoav=0.25*(uri(1,i,j,k)+uri(1,i-1,j,k) &
                     +uri(1,i,j,k-1)+uri(1,i-1,j,k-1))
        rhoav=uri(1,i,j,k)
!
        if(rhoav .ne. 0.d0) then
          pom1=rhoav/rhomax1a - 0.3
        else
          pom1=0.d0
        endif
!
!         if(j .eq. 4 .and. k .eq. 64) then
!           write(*,*) 'pom1=', pom1, i, j, k
!         endif
!
        if(pom1 .gt. 0.d0) then
          pom(i,j,k)=pom1
        endif
!
      enddo
    enddo
  enddo
!
!- Set-up of Magnetic field -!
  do k=ks1+1,ke1-1
    do j=js1,je1
      do i=is1+1,ie1-1
!        
!!        bcon(1)=-(pom(i,j,k)-pom(i,j,k+1)+pom(i+1,j,k)-pom(i+1,j,k+1)) &
!!                   /(2.*(x3(k+1)-x3(k))*detg(i,j,k))
!!
!!        bcon(3)=(pom(i,j,k)+pom(i,j,k+1)-pom(i+1,j,k)-pom(i+1,j,k+1)) &
!!                   /(2.*(x1(i+1)-x1(i))*detg(i,j,k))
!
        tmp1a=pom(i,j,k)-pom(i,j,k+1)+pom(i+1,j,k)-pom(i+1,j,k+1)
        if(tmp1a .ne. 0.d0) then
          bcon(1)=-tmp1a/(2.*dx3(k)*detg(i,j,k))
        else
          bcon(1)=0.d0
        endif
!!
        tmp1b=pom(i,j,k)+pom(i,j,k+1)-pom(i+1,j,k)-pom(i+1,j,k+1)
        if(tmp1b .ne. 0.d0) then
          bcon(3)=tmp1b/(2.*dx1(i)*detg(i,j,k))
        else
          bcon(3)=0.d0
        endif
!
!        if(k .eq. ks1) then
!          bcon(1)=(pom(i,j,k+1)-pom(i,j,k))/((x3(k+1)-x3(k))*detg(i,j,k))
!        elseif(k .eq. ke1) then
!          bcon(1)=(pom(i,j,k)-pom(i,j,k-1))/((x3(k)-x3(k-1))*detg(i,j,k))
!        else         
!          bcon(1)=(pom(i,j,k+1)-pom(i,j,k-1))/((x3(k+1)-x3(k-1))*detg(i,j,k))
!        endif
!
!        if(i .eq. is1) then
!          bcon(3)=-(pom(i+1,j,k)-pom(i,j,k))/((x1(i+1)-x1(i))*detg(i,j,k))
!        elseif(i .eq. ie1) then
!          bcon(3)=-(pom(i,j,k)-pom(i-1,j,k))/((x1(i)-x1(i-1))*detg(i,j,k))
!        else         
!          bcon(3)=-(pom(i+1,j,k)-pom(i-1,j,k))/((x1(i+1)-x1(i-1))*detg(i,j,k))
!        endif
!
        uri(7,i,j,k)=bcon(1)
        uri(9,i,j,k)=bcon(3)
!
      enddo
    enddo
  enddo
!
!- put boundary data for magnetic field -!
!
  do j=js1,je1
    do i=is1,ie1
      uri(7,i,j,ks1)=uri(7,i,j,ks1+1)
      uri(7,i,j,ke1)=uri(7,i,j,ke1-1)
!
      uri(9,i,j,ks1)=uri(9,i,j,ks1+1)
      uri(9,i,j,ke1)=uri(9,i,j,ke1-1)
    enddo
  enddo
!
  do k=ks1,ke1
    do j=js1,je1
      uri(7,is1,j,k)=uri(7,is1+1,j,k)
      uri(7,ie1,j,k)=uri(7,ie1-1,j,k)
!
      uri(9,is1,j,k)=uri(9,is1+1,j,k)
      uri(9,ie1,j,k)=uri(9,ie1-1,j,k)
    enddo
  enddo
!
!- calculation maximum magnetic field strength -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        util(1)=uri(2,i,j,k)
        util(2)=uri(3,i,j,k)
        util(3)=uri(4,i,j,k)
        bcon(1)=uri(7,i,j,k)
        bcon(2)=uri(8,i,j,k)
        bcon(3)=uri(9,i,j,k)
!       
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
          enddo
        enddo
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(bbsq .gt. bsq_max) then
          bsq_max=bbsq
        endif
!
      enddo
    enddo
  enddo    
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(bsq_max,bsq_max1,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
!
  beta_act=prmax2a/(0.5*bsq_max1)
!
  if(beta .eq. 0.d0) then
    bnorm=0.d0
  else
    bnorm=sqrt(beta_act/beta)
  endif
!
  if(myrank .eq. 0) then
    write(*,*) "beta_act, bnorm=", beta_act, bnorm
  endif
!
!- Normalize magnetic field strength -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        bcon(1)=uri(7,i,j,k)*bnorm            
        bcon(2)=uri(8,i,j,k)*bnorm  
        bcon(3)=uri(9,i,j,k)*bnorm  
!
        util(1)=uri(2,i,j,k)
        util(2)=uri(3,i,j,k)
        util(3)=uri(4,i,j,k)
!
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)
        uri(9,i,j,k)=bcon(3)
!
        do m=0,3
          do n=0,3
            gcov1(m,n)=gcov(m,n,i,j,k)
          enddo
        enddo        
!
        call calgfl(util,gfl,gcov1)
        call calucon(util,ucon,gfl,gcon1)
        call lower(ucon,ucov,gcov1)
        call calbbcon(bcon,bbcon,ucov,ucon)
        call lower(bbcon,bbcov,gcov1)
        call calbbsq(bbcon,bbcov,bbsq)
!
        if(bbsq .gt. bsq_max2) then
          bsq_max2=bbsq
        endif
!
      enddo
    enddo
  enddo 
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(bsq_max2,bsq_max2a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  beta_act=prmax2a/(0.5*bsq_max2a)
!
  if(myrank .eq. 0) then
    write(*,*) "beta_act=", beta_act
  endif    
!
  deallocate(gcov1, gcon1, gcov1in, util, utiln, bcon, ucov, ucon, &
             bbcov, bbcon, beta1, ah, pom, stat=merr)
!
  return
end subroutine mdtorus3
!
!--------------------------------------------------------------------
subroutine mdrecoilbh(uri,gcov,gcon,detg, &
                      x1,x2,x3,x1b,x2b,x3b,dx1,dx2,dx3, &
                      myrank,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!  Recoiling BHs, Zanotti et al. (2010)
!  model=10 in pram.f90
!
  use pram, only : imax, jmax, kmax, nv, gam, metric, akm, pi, iter, kpol
  implicit none
  include 'mpif.h'

  integer :: i, j, k, m, n, is1 ,ie1, js1, je1, ks1, ke1
  integer :: merr, myrank, jh, nnn

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)
  
  real(8) :: gcov(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: gcon(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
!
  real(8) :: detg(is1:ie1,js1:je1,ks1:ke1)
!
  real(8), allocatable :: gcov1(:,:), gcon1(:,:), gcov1in(:,:)
  real(8), allocatable :: ucov(:), ucon(:), bbcov(:), bbcon(:), beta1(:)
  real(8), allocatable :: util(:), utiln(:), bcon(:)
!
  real(8) :: x1(imax), x2(jmax), x3(kmax) 
  real(8) :: x1b(imax), x2b(jmax), x3b(kmax) 
  real(8) :: dx1(imax), dx2(jmax), dx3(kmax)
!
  real(8),allocatable :: ah(:,:,:)
!
  real(8) :: de, pr  
  real(8) :: kappa, beta, clm, utin, ut, uut, &
             rhomin, prmin, hm1, omg, &
             rhomax1, rhomax1a, prmax1, prmax1a, prmax2, prmax2a 
  real(8) :: rr, ph, th, sth, cth, del, aa, sig, rr1, th1, &
             rin, thin, sthin, cthin, delin, aain, sigin, rin1
  real(8) :: x1aa, x2aa, x3aa, x3t
  real(8) :: tmp_rand
  real(8) :: bsq_max, bsq_max1, bsq_max2, bsq_max2a, beta_act, &
             bnorm, pom1, rhoav, gfl, bbsq
  real(8) :: tmp4a, tmp4b, tmp4c, discr, alpha1
  real(8) :: tmp1a, tmp1b, tmp2a, tmp2b
  real(8) :: ucovph
  real(8) :: vrecoil
! 
!- allocate variables -!
  allocate(gcov1(0:3,0:3), gcon1(0:3,0:3), gcov1in(0:3,0:3), stat=merr)
  allocate(util(1:3), utiln(1:3), bcon(1:3), ucov(0:3), ucon(0:3), &
           bbcov(0:3), bbcon(0:3), beta1(1:3), stat=merr)
  allocate(ah(is1:ie1,js1:je1,ks1:ke1), stat=merr)

!--------------------------------------------------------------------
!- Parameter -!

  rin=40.0d0
!  rin=9.0d0
  thin=pi/2.d0
  kappa=1.0d-3
  beta=0.d0
!  beta=100.d0
!  beta=0.d0
!  clm=4.5d0
  clm=8.0d0
!  
  rin1=1.d0
  rhomin=1.d-6 
  prmin=1.d-3*rhomin
!
  rhomax1=0.d0
  rhomax1a=0.d0
  prmax1=0.d0
  prmax1a=0.d0
  prmax2=0.d0
!
  vrecoil=0.001d0
!  
!- initally set minimum plasma everywhere -!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1

        do m=0,3
          do n=0,3
            gcon1(m,n)=gcon(m,n,i,j,k)
          enddo
        enddo

        rr=x1(i)

        de=rhomin
        pr=prmin
        util(1)=0.d0
!        util(1)=(gcon1(0,1)/gcon1(0,0))*(1.-(1./rr)**4)
        util(2)=0.d0
        util(3)=0.d0
        bcon(1)=0.d0
        bcon(2)=0.d0 
        bcon(3)=0.d0
!          
!- set primitive variables -!
!          
        uri(1,i,j,k)=de
        uri(2,i,j,k)=util(1)
        uri(3,i,j,k)=util(2)
        uri(4,i,j,k)=util(3)
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)          
        uri(9,i,j,k)=bcon(3)
!       
      enddo
    enddo
  enddo
!  
!- calculation of BL Kerr metric at rin, thin -!
!  
  sthin=sin(thin)
  cthin=cos(thin)
  delin=rin*rin-2.*rin+akm*akm
  aain=(rin*rin+akm*akm)*(rin*rin+akm*akm)-delin*akm*akm*sthin*sthin
  sigin=rin*rin+akm*akm*cthin*cthin  
!
  do n=0,3
    do m=0,3
      gcov1in(m,n)=0.d0
    enddo
  enddo

  gcov1in(0,0)=-(sigin-2.*rin)/sigin               !- g_tt -!
  gcov1in(1,1)=sigin/delin                          !- g_rr -!
  gcov1in(2,2)=(aain/sigin)*sthin*sthin !- g_phi phi -!
  gcov1in(0,2)=-(2.*akm*rin*sthin*sthin)/sigin  !- g_t phi -!
  gcov1in(2,0)=gcov1in(0,2)                  !- g_phi t -!
  gcov1in(3,3)=sigin                      !- g_th th -!
! 
!- calculation of ut at rin -!
  tmp1a=gcov1in(0,2)*gcov1in(0,2)-gcov1in(0,0)*gcov1in(2,2)
  tmp1b=gcov1in(0,0)*clm*clm+2.*clm*gcov1in(0,2)+gcov1in(2,2)
  if(tmp1b .gt. 0.d0) then
    utin=-sqrt(tmp1a/tmp1b)
  else
    utin=-1.d0  
  endif
!
  if(myrank .eq. 0) then
    write(*,*) 'lin, utin=', clm, utin
  endif
!
!- calculate h -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!
        rr=x1(i)
         
        if(x2(j) .lt. 0.d0) then
          ph=2.*pi-x2(j)
        elseif(x2(j) .gt. 2.*pi) then
          ph=x2(j)-2.*pi
        else   
          ph=x2(j) 
        endif

!        if(i .eq. 10 .and. k .eq. 4) then
!          write(*,*) "j, x2, ph=", j, x2(j), ph
!        endif   
        
        th=x3(k)
        sth=sin(th)
        cth=cos(th) 
!
        if(rr .ge. rin) then         
!- calculation of BL Kerr metric -! 
          del=(rr*rr)-(2.*rr)+(akm*akm)
          aa=(rr*rr+akm*akm)*(rr*rr+akm*akm)-del*akm*akm*sth*sth
          sig=rr*rr+akm*akm*cth*cth
!
          do n=0,3
            do m=0,3
              gcov1(m,n)=0.d0
              gcon1(m,n)=0.d0 
            enddo
          enddo
!
          gcov1(0,0)=-(sig-2.*rr)/sig               !- g^tt -!
          gcov1(1,1)=sig/del                          !- g^rr -!
          gcov1(2,2)=(aa/sig)*sth*sth             !- g^phi phi -!
          gcov1(0,2)=-(2.*akm*rr*sth*sth)/sig           !- g^t phi -!
          gcov1(2,0)=gcov1(0,2)                  !- g^phi t -!
          gcov1(3,3)=sig                      !- g^th th -!        
!
          gcon1(0,0)=-aa/(sig*del)               !- g^tt -!
          gcon1(1,1)=del/sig                          !- g^rr -!
          gcon1(2,2)=(sig-2.*rr)/(del*sig*sth*sth) !- g^phi phi -!
          gcon1(0,2)=-(2.*akm*rr)/(sig*del)           !- g^t phi -!
          gcon1(2,0)=gcon1(0,2)                  !- g^phi t -!
          gcon1(3,3)=1./sig 
!          
!- calculation of u_t at r,theta -!
!          
          tmp2a=gcov1(0,2)*gcov1(0,2)-gcov1(0,0)*gcov1(2,2)
          tmp2b=gcov1(0,0)*clm*clm+2.*clm*gcov1(0,2)+gcov1(2,2)
          if(tmp2b .gt. 0.d0) then
            ut=-sqrt(tmp2a/tmp2b)
          else
            ut=-1.d0  
          endif        

!
!- calculation of h = u_t,in/u_t
!          
          ah(i,j,k)=utin/ut
!
        else
          ah(i,j,k)=-1.d0
        endif
!
!- set torus structure -!
!
        if(ah(i,j,k) .gt. 1.d0 .and. rr .ge. rin) then    
!    
          hm1=max((ah(i,j,k) -1.)*(gam-1.)/gam, 0.d0)
          de=(hm1/kpol)**(1./(gam-1.))
          pr=kpol*(hm1/kpol)**(gam/(gam-1.))
!
!         if(j .eq. 4 .and. k .eq. 32) then
!           write(*,*) 'ah, de=', ah(i,j,k), de(i,j,k), i
!         endif
!
          util(1)=0.d0
          util(3)=0.d0
!
!- calculation of omega & contravariant u^t
!          
          omg=-(gcov1(0,2)+gcov1(0,0)*clm)/(gcov1(2,2)+gcov1(0,2)*clm)          
          uut=sqrt(-1./(gcov1(0,0)+2.*omg*gcov1(0,2)+omg*omg*gcov1(2,2)))
!          uut=-1./(ut*(1.-omg*clm))
!
          ucovph=-omg*ut

          util(2)=(omg*uut)
!          util(2)=gcon1(0,2)*ut+gcon1(2,2)*ucovph
!          util(2)=0.d0
!
!- Put perturbation in pressure -!
!
!         call random_number(tmp_rand)
!!          write(*,*) "tmp_rand=", tmp_rand-0.5
!          pr=pr*(1.+0.05*(tmp_rand-0.5))
!
!- Put recoiling speed -!
!
          util(1)=util(1)+vrecoil*cos(ph)
          util(2)=util(2)-vrecoil*sin(ph)

          
!- check density & pressure max -!
!
          if(de .gt. rhomax1) then
            rhomax1=de
!            write(*,*) 'de,i,k,rr,th=', de,i,k,x1(i),x3(k)
          endif
          if(pr .gt. prmax1) then
            prmax1=pr
          endif         
!
!- transform velocity from BL to KS coordinates -!
!
          if(metric .eq. 303) then
            x1aa=x1(i)
            x2aa=x2(j)
            x3aa=x3(k)   
            x3t=x3b(k)
            call transbl2ks(util,utiln,x1aa,x2aa,x3aa,x3t)
            util(1)=utiln(1)
            util(2)=utiln(2)
            util(3)=utiln(3)
!
          elseif(metric .eq. 203 .and. akm .ne. 0.d0) then
!
            do m=0,3
              do n=0,3
                gcov1(m,n)=gcov(m,n,i,j,k)
                gcon1(m,n)=gcon(m,n,i,j,k)
              enddo
            enddo
!
            alpha1=1./sqrt(-gcon1(0,0))
!
            beta1(1)=alpha1*alpha1*gcon1(0,1)
            beta1(2)=alpha1*alpha1*gcon1(0,2)
            beta1(3)=alpha1*alpha1*gcon1(0,3)
!
            ucon(1)=util(1)
            ucon(2)=util(2)
            ucon(3)=util(3)
!
            tmp4a=gcov1(0,0)
            tmp4b=2.*(gcov1(0,1)*ucon(1)+gcov1(0,2)*ucon(2)+gcov1(0,3)*ucon(3))
            tmp4c=1.+gcov1(1,1)*ucon(1)*ucon(1)+gcov1(2,2)*ucon(2)*ucon(2) &    
                  +gcov1(3,3)*ucon(3)*ucon(3) &
                  +2.*(gcov1(1,2)*ucon(1)*ucon(2)+gcov1(1,3)*ucon(1)*ucon(3) &  
                  +gcov1(2,3)*ucon(2)*ucon(3))   
            discr=tmp4b*tmp4b-4.*tmp4a*tmp4c
            ucon(0)=(-tmp4b-sqrt(discr))/(2.*tmp4a)
! 
            gfl=ucon(0)*alpha1
!
            util(1)=ucon(1)+gfl*beta1(1)/alpha1
            util(2)=ucon(2)+gfl*beta1(2)/alpha1
            util(3)=ucon(3)+gfl*beta1(3)/alpha1  
!
          endif
!
!          
!- set primitive variables -!
!          
          uri(1,i,j,k)=de
          uri(2,i,j,k)=util(1)
          uri(3,i,j,k)=util(2)
          uri(4,i,j,k)=util(3)
          uri(5,i,j,k)=pr
!          
        endif
!
      enddo
    enddo
  enddo
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_allreduce(rhomax1,rhomax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)
  call mpi_allreduce(prmax1,prmax1a,1,mpi_double_precision,mpi_max, &
                     mpi_comm_world,merr)

  if(myrank .eq. 0) then
    write(*,*) "rhomax1a, prmax1a=", rhomax1a, prmax1a
  endif
!
!
!- Normalize the densities and pressure -!
!
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1 
        rr=x1(i)
        th=x3(k)
        if(ah(i,j,k) .gt. 1.d0 .and. rr .ge. rin) then
          de=uri(1,i,j,k)/rhomax1a
          pr=uri(5,i,j,k)/rhomax1a
!
          if(de .le. rhomin) then
            de=rhomin
          endif
          if(pr .le. prmin) then
            pr=prmin
          endif
!
          uri(1,i,j,k)=de
          uri(5,i,j,k)=pr
        endif
!
      enddo
    enddo
  enddo
!
   prmax2a=prmax1a/rhomax1a
   rhomax1a=1.d0
!
  if(myrank .eq. 0) then
    write(*,*) 'rhomax1a, prmax2a=', rhomax1a, prmax2a
  endif   
!
!- put recoiling BH speed -!
!  
!  do k=ks1,ke1
!    do j=js1,je1
!      do i=is1,ie1
!          
!        uri(2,i,j,k)=uri(2,i,j,k)+vrecoil
!         
!      enddo
!    enddo  
!  enddo
!         
  deallocate(gcov1, gcon1, gcov1in, util, utiln, bcon, ucov, ucon, &
             bbcov, bbcon, beta1, ah, stat=merr)
!
  return
end subroutine mdrecoilbh
!
!--------------------------------------------------------------------
subroutine mdwindcol(uri,gcov,gcon,x1,x2,x3, &
                     myrank,is1,ie1,js1,je1,ks1,ke1)
!--------------------------------------------------------------------
!         model = 11; 2D Wind Collision Problem

  use pram, only : nv, imax, jmax, kmax, gam, c0, pi
  implicit none
  include 'mpif.h' ! for MPI
  
  integer :: i, j, k, n, m, is1, ie1, js1, je1, ks1, ke1
  integer :: myrank, merr 

  integer, parameter :: nkmax=8

  real(8) :: uri(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: gcov(0:3,0:3,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: gcon(0:3,0:3,is1:ie1,js1:je1,ks1:ke1) 
!
  real(8) :: x1(imax), x2(jmax), x3(kmax)
!
  real(8), allocatable :: gcov1(:,:), gcon1(:,:)
  real(8), allocatable :: vel(:), bcon(:), phi1(:)
! 
  real(8) :: rr, th, de, pr, rsh, tmp1, tmp2, tmp3, vsq, gfl, gflmax
!
!-allocate variables -!
  allocate(gcov1(0:3,0:3), gcon1(0:3,0:3), stat=merr)
  allocate(vel(1:3), bcon(1:3), phi1(1:nkmax), stat=merr)  
  
!--------------------------------------------------------------------
!- Parameter -!
!
  rsh=1000.d0
  gflmax=10.d0

  if(myrank .eq. 0) then
    do n=1, nkmax
      call random_number(tmp3)
      phi1(n)=2.0*pi*tmp3
!      phi1(n)=0.d0       
    enddo
  endif
!
  call mpi_barrier(mpi_comm_world,merr)
  call mpi_bcast(phi1,nkmax,mpi_double_precision,0,mpi_comm_world,merr)    
!    
!
! Initial setup
!  
  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
!         
        rr=x1(i)
        th=x3(k)
!        
        if(rr .le. rsh) then
!- fast wind region
          de=0.01d0 
          pr=0.001d0
!          vcon(1)=0.4d0
          vel(1)=0.d0
          vel(2)=0.d0
          vel(3)=0.d0
          bcon(1)=0.d0
          bcon(2)=0.d0
          bcon(3)=0.d0
           
        elseif(rr .gt. rsh) then
!- slow wind region           
          de=1.d0
          pr=0.01d0
!          pr=0.001d0
!          vel(1)=0.1d0
          vel(1)=0.d0
          vel(2)=0.d0
          vel(3)=0.d0
          bcon(1)=0.d0
          bcon(2)=0.d0
          bcon(3)=0.d0
!
        endif
!
!- perturbation
!       
!- random number -!
!        call random_number(tmp1)
!        tmp2=2.0*tmp1-1.0
!- sinusoidal -!
        do n=1, nkmax
          tmp2=tmp2+sin(float(n)*th+phi1(n))
        enddo   
        tmp2=tmp2/float(nkmax)  
!       
        if(rr .gt. 0.9d0*rsh .and. rr .lt. 1.1d0*rsh) then
          vel(1)=vel(1)+0.05*tmp2 
        endif
!      
!- convert 3-vel => 4-vel
!
        vsq=0.d0
        do m=1,3
          do n=1,3
            vsq=vsq+gcov1(m,n)*vel(m)*vel(n) 
          enddo
        enddo
!
        if(vsq .ge. 1.d0) then
          vsq=1.-1./(gflmax*gflmax)
        endif   
!- new set of primitive variables -!
        gfl=1./sqrt(1.-vsq)
        uri(1,i,j,k)=de
        uri(2,i,j,k)=vel(1)*gfl
        uri(3,i,j,k)=vel(2)*gfl
        uri(4,i,j,k)=vel(3)*gfl
        uri(5,i,j,k)=pr
        uri(6,i,j,k)=0.d0
        uri(7,i,j,k)=bcon(1)
        uri(8,i,j,k)=bcon(2)
        uri(9,i,j,k)=bcon(3)
!        
      enddo
    enddo
  enddo
!
  deallocate(gcov1, gcon1, vel, bcon, phi1, stat=merr)
!
  return
end subroutine mdwindcol
!
!***********************************************************************
!           GENERAL FUNCTIONS
!***********************************************************************
!--------------------------------------------------------------------
function ffjump(xx,bi,ee,aa,dd,bb)
!--------------------------------------------------------------------
  implicit none
  real(8) :: xx, bi, ee, aa, dd, bb, ffjump
!
  if( xx .le. aa-bb ) then
    ffjump=bi
   elseif( xx .ge. aa+bb ) then
    ffjump=ee
   else
    ffjump=1.0+tanh((xx-aa)/dd)/tanh(bb/dd)
    ffjump=bi+0.5*ffjump*(ee-bi)
   endif
!
   return
end function ffjump
!

